﻿//xjasek04
//Bachelor thesis - Physical Simulation in 3D Scene Using PhysX

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Graphics;
using PhysXTanks.Input;

namespace PhysXTanks
{
    public class Camera : KeyboardListener, MouseListener
    {
        #region Variables

        private GraphicsDevice device;
        private GameWindow window;

        private Matrix _view, _projection;
        private Matrix cameraRotation;
        private float _cameraPitch, _cameraYaw;
        private Vector3 cameraTarget;
        private Vector3 newPosition;
        private Vector3 newForward;

        private float elapsed;
        private Game game;

        private Boolean debug = false;
        #endregion


        internal Camera(GraphicsDevice device, GameWindow window)
        {
            game = new Game();
            this.device = device;
            this.window = window;                                                            
            this.View = Matrix.CreateLookAt(new Vector3(0.0f, 50.0f, -80.0f), new Vector3(0.0f, 50.0f, -75.0f), new Vector3(0.0f, 1.0f, 0.0f));
            this.Projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, device.Viewport.AspectRatio, 0.5f, 100000.0f);            
            CenterCursor();
            InputManager.Instace.AddKeyboardListener(this);
            InputManager.Instace.AddMouseListener(this);        
        }

        public void Update(GameTime elapsedTime)
        {
            //elapsed time since last frame in seconds
            elapsed = (float)(elapsedTime.ElapsedGameTime.TotalMilliseconds / 1000.0);

            if (debug == true)
                return;

            Vector3 camUp = Vector3.Up;
            camUp = Vector3.Transform(camUp, Matrix.CreateFromQuaternion(Game1.TankOrientationQuat));
            Vector3 cameraPos = new Vector3(0.0f, 25.0f, 80.0f); //tank position 60,0,0      
            cameraPos = Vector3.Transform(cameraPos, Matrix.CreateFromQuaternion(Game1.TankOrientationQuat));
            cameraPos += Game1.TankPosition;

            _view = Matrix.CreateLookAt(cameraPos, Game1.TankPosition, Vector3.Up); 
        }

        private void CenterCursor()
        {
            Mouse.SetPosition(window.ClientBounds.Width / 2, window.ClientBounds.Height / 2);
        }

        #region Public Properties

        public Matrix View
        {
            get
            {
                return _view;
            }
            set
            {
                this._view = value; 
            }
        }

        public Matrix Projection
        {
            get
            {
                return _projection;
            } 
            set
            {
                this._projection = value;
            }
        }

        public Vector3 CameraTarget
        {
            get
            {
                return cameraTarget;
            }
        }

        public float CameraYaw
        {
            get
            {
                return _cameraYaw;
            }

            set
            {
                this._cameraYaw = value;
            }
        }

        public float CameraPitch
        {
            get
            {
                return _cameraPitch;
            }
            set
            {
                this._cameraPitch = value;
            }
        }

        #endregion

        #region KeyboardListener Members

        public void KeyPressed(Keys key)
        {            
        }

        public void KeyPressedRunning(Keys key)
        {
        }

        public void KeyHoldedRunning(Keys key)
        {
            if (debug == false)
                return;
            // The amount of movement * the direction of movement, then rotate that along the direction we are looking
            Vector3 translateDirection = Vector3.Zero;
            Vector3 position = Matrix.Invert(this.View).Translation;

            switch (key)
            {
                case Keys.W: // Forwards
                    translateDirection += Vector3.TransformNormal(Vector3.Forward, cameraRotation) * 5.0f;
                    break;
                case Keys.S: // Backwards
                    translateDirection += Vector3.TransformNormal(Vector3.Backward, cameraRotation) * 5.0f;
                    break;
                case Keys.A: // Left
                    translateDirection += Vector3.TransformNormal(Vector3.Left, cameraRotation) * 5.0f;
                    break;
                case Keys.D: // Right
                    translateDirection += Vector3.TransformNormal(Vector3.Right, cameraRotation) * 5.0f;
                    break;
            }

            const float speed = 100.0f; // 100 distance units per second
            float distance = speed * elapsed; // d = v*t  

            newPosition = position;
            if (translateDirection.LengthSquared() > 0)
            {
                newPosition += Vector3.Normalize(translateDirection) * distance;
            }

            newForward = Vector3.TransformNormal(Vector3.Forward, cameraRotation);

            if (CurrentState.Instance.GameState == GameState.Running)
            {
                _view = Matrix.CreateLookAt(newPosition, newPosition + newForward, Vector3.Up);              
            }
            cameraTarget = new Vector3(newPosition.X + newForward.X, newPosition.Y + newForward.Y,
                                       newPosition.Z + newForward.Z); 
        }

        public void KeyPressedInMenu(Keys key)
        {
        }

        #endregion

        #region MouseListener Members
        public void MouseMoved(MouseState mouseState)
        {
            if (debug == false)
                return;

            Vector2 cursorPosition = new Vector2(mouseState.X, mouseState.Y);
            Vector2 mouseCenter = new Vector2(window.ClientBounds.Width / 2, window.ClientBounds.Height / 2);
            Vector2 delta = cursorPosition - mouseCenter;
            Vector2 deltaDampened = delta * 0.0015f;
            if (CurrentState.Instance.GameState == GameState.Running && this.game.IsActive)
            {
                CenterCursor();
                //camera Yaw and Pitch extraction
                _cameraYaw -= deltaDampened.X;
                _cameraPitch -= deltaDampened.Y;
            }
            else if (CurrentState.Instance.GameState == GameState.InMenu)
            {
                _cameraYaw = this.CameraYaw;
                _cameraPitch = this.CameraPitch;
            }

            Vector3 position = Matrix.Invert(this.View).Translation;

            cameraRotation = Matrix.CreateFromYawPitchRoll(_cameraYaw, _cameraPitch, 0.0f);
            newForward = Vector3.TransformNormal(Vector3.Forward, cameraRotation);

            if (CurrentState.Instance.GameState == GameState.Running)
            {
                _view = Matrix.CreateLookAt(newPosition, newPosition + newForward, Vector3.Up);             
            }
            cameraTarget = new Vector3(newPosition.X + newForward.X, newPosition.Y + newForward.Y,
                                       newPosition.Z + newForward.Z);
        }

        public void MouseLeftButtonPressed()
        {
        }

        public void MouseRightButtonPressed()
        {
        }
        #endregion               
    }
}