//xjasek04
//Bachelor thesis - Physical Simulation in 3D Scene Using PhysX

#define NETWORK

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
    
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

using StillDesign.PhysX;

using PhysXTanks;
using System.IO;
using PhysXTanks.Input;
using PhysXTanks.Entities;
using Lidgren.Network;

namespace PhysXTanks
{
    enum CollisionGroup
    {
        GROUP_ONE,
        GROUP_TWO,
        GROUP_THREE,
        GROUP_NON_COLLIDABLE,
        GROUP_COLLIDABLE_NON_PUSHABLE,
        GROUP_COLLIDABLE_PUSHABLE,
    };

    enum messageType 
    { 
        POSITION, 
        BULLET_DESTROYED,
        HIT
    };

    public class Game1 : Microsoft.Xna.Framework.Game, KeyboardListener, MouseListener
    {
        private SpriteBatch spriteBatch;
        private GraphicsDeviceManager deviceManager;
        private BasicEffect VisualizationEffect;
        private float elapsedTimeFloat;        
        private double elapsedTotalTime;
        private double activeTime;
        
        Vector3 gForceVec = new Vector3(0, 0, 0);
        float gForceStrenght = 20000;

        private const int DEBUG_MODE_STR = 1;
        
        KeyboardState keyboardState = Keyboard.GetState();

        //camera properties and auxialiry variables
        Camera Camera;
        private float cameraYaw;
        private float cameraPitch;        
        //<end tag>
        
        Core Core;
        Scene Scene;

        bool isFilteringOn = false;        

        bool bIsDebug;
        bool forceKeysDisabled;

        //FPS
        private int frameRate = 0;
        private int frameCounter = 0;
        TimeSpan elapsedTime = TimeSpan.Zero;

        private bool markSpleepObjects;
        private Vector3 defaultColor;
        private Vector3 selectedColor;
        private Vector3 tankColor;

        SpriteFont font;
        Vector2 FontPos;

        Sphere sphere = new Sphere(10.0f);
        Plane groundPlane = new Plane();
        Box worldBox;

        //game mode
        private Boolean multiplayer; 

        //skybox
        Texture2D[] skyboxTextures;
        Model skyboxModel;
        Effect effect;

        //plane texture
        private Texture2D planeTexture;

        //health
        private Texture2D lifeBarOrigin;
        private Texture2D lifeBar;

        //reload
        private Texture2D reloadBarOrigin;
        private Texture2D reloadBar;
        private Int16 reloadProgress;
        private Boolean reloadActive;
        private Boolean reloadActivate;

        //models
        private Model _sphere;
        private Model _box;
        private Model _tank;
        private Model _capsule;
        private Model _plane;

        //actors
        private static Actor tank;
        private Actor gSelectedActor;       
        private Actor bullet;
        private Actor enemyActor;
        private Actor box0;         

        //command line arguments
        private static string[] parameters;

        //tank movement - model rotation, forward vector and backward vector for
        //movement 
        private Quaternion _modelRotation;
        private Quaternion rot;
        private Vector3 newForward;
        private Vector3 newBackward;

        //Network  properties 
        private static Dictionary<long, List<Matrix>> positions2;
        private List<Matrix> positionsMatrixes;
        private Matrix matrixPose = new Matrix();
        private NetClient client;
        private Dictionary<long, Tank> enemyTanks;
        private long playerID = 0;
        //Network

        //list of enemies
        private Dictionary<long, Actor> enemyActorsList;

        //sound
        //Set the sound effects to use
        private SoundEffect soundExplosion;
        private SoundEffect soundReload;   
        private SoundEffect soundFire;
        private SoundEffectInstance soundExplosionInstance;
        private SoundEffectInstance soundReloadInstance;
        private SoundEffectInstance soundFireInstance;
        private Boolean soundMuted;

        //music
        private Song gameMusic;

        //Collisions
        private double collisionCountdown;
        private Boolean enemyHit = false;
        //private Boolean collisionOccured;
        //Collisions

        //bullets
        private Boolean bulletShotted;
        private Dictionary<int, Actor> bulletsList;
        private int i = 0;
        private int bulletsKeys = 0;
        private Dictionary<int, double> bulletsDestroyList; //int = bullet ID, double = time to destroy
        private Boolean destroyEnemyBullet = false;
        //list of enemy bullets
        private Actor enemyBulletActor;
        private Matrix matrixBulletPose = new Matrix();

        private Vector3 tankSize;

        private bool gameStarted = false;

        Tank myTank = new Tank();        
        private Boolean debug = true;
        

        //cloth
        List<Cloth> cloths = new List<Cloth>();
        private Cloth _cloth;


        public Game1()
        {
            deviceManager = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            // Add to KeyboardListeners HashSet
            InputManager.Instace.AddKeyboardListener(this);
            InputManager.Instace.AddMouseListener(this);

            //Networking
            #if NETWORK
            NetPeerConfiguration config = new NetPeerConfiguration("xnaapp");
            config.EnableMessageType(NetIncomingMessageType.DiscoveryResponse);

            client = new NetClient(config);
            client.Start();
            #endif
        }

        protected override void Initialize()
        {            
            if (parameters.Length != 0 &&  parameters[0] == "multiplayer")
                multiplayer = true;
            else
                multiplayer = false;

            bIsDebug = true;
            forceKeysDisabled = false;
            
            soundMuted = false;
            MediaPlayer.IsRepeating = true;


            positions2 = new Dictionary<long,List<Matrix>>();
            positionsMatrixes = new List<Matrix>();

            //new stuff make it look nice - reorganize it you moron!
            tankColor = RGBToColor(new Vector3(105.0f, 105.0f, 105.0f)); //dark grey
            defaultColor = RGBToColor(new Vector3(220.0f, 220.0f, 220.0f)); //light grey
            selectedColor = RGBToColor(new Vector3(0.0f, 255.0f, 0.0f)); //green

            //tank positioning - for shooting and movement
            newForward = new Vector3(-3.724223E-10f, 0.0f, -1.0f);
            newBackward = new Vector3(3.724223E-10f, 0.0f, 1.0f);

            client.DiscoverLocalPeers(14242);

            this.Camera = new Camera(GraphicsDevice, Window);

            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);            

            VisualizationEffect = new BasicEffect(GraphicsDevice, null);
            VisualizationEffect.VertexColorEnabled = true;          

            VisualizationEffect.DiffuseColor = new Vector3(1.0f, 3.0f, 10.0f);

            //initPhysx()
            this.Core = new Core();
            var core = this.Core;

            #region Core Parameters

            core.SetParameter(PhysicsParameter.VisualizationScale, 1.0f);
            core.SetParameter(PhysicsParameter.VisualizeCollisionShapes, true);

            core.SetParameter(PhysicsParameter.VisualizeClothMesh, true);
            core.SetParameter(PhysicsParameter.VisualizeClothCollision, true);
            core.SetParameter(PhysicsParameter.VisualizeClothSelfCollision, true);
            core.SetParameter(PhysicsParameter.VisualizeClothWorkPackets, true);
            core.SetParameter(PhysicsParameter.VisualizeClothSleep, true);
            core.SetParameter(PhysicsParameter.VisualizeClothSleepVertex, true);
            core.SetParameter(PhysicsParameter.VisualizeClothTearableVertices, true);
            core.SetParameter(PhysicsParameter.VisualizeClothTearing, true);
            core.SetParameter(PhysicsParameter.VisualizeClothAttachment, true);                                                           
            core.SetParameter(PhysicsParameter.VisualizeClothValidBounds, true);            

            core.SetParameter(PhysicsParameter.VisualizeFluidPosition, false);
            core.SetParameter(PhysicsParameter.VisualizeFluidEmitters, false); // Slows down rendering a bit too much
            core.SetParameter(PhysicsParameter.VisualizeForceFields, true);
            core.SetParameter(PhysicsParameter.VisualizeSoftBodyMesh, false);
            core.SetParameter(PhysicsParameter.DefaultSleepLinearVelocitySquared, 2.0f * 2.0f);
            core.SetParameter(PhysicsParameter.DefaultSleepAngularVelocitySquared, 2.0f * 2.0f);
            core.SetParameter(PhysicsParameter.ContinuousCollisionDetection, true);

            core.SetParameter(PhysicsParameter.SkinWidth, 0.025f);
            core.SetParameter(PhysicsParameter.VisualizeActorAxes, true);            

            #endregion

            SceneDescription sceneDesc = new SceneDescription();
            sceneDesc.SimulationType = SimulationType.Software;
            sceneDesc.Gravity = new Vector3(0.0f, -9.81f, 0.0f);
            sceneDesc.GroundPlaneEnabled = true;
           
            this.Scene = core.CreateScene(sceneDesc); 

            Material defaultMat = this.Scene.Materials[0];
                
            //box around the scene
            worldBox = new Box(new Vector3(0.0f, 0.0f, 0.0f), new Vector3(0.0f, 0.0f, 0.0f));

            defaultMat.Restitution = 0.5f;
            defaultMat.StaticFriction = 0.5f;
            defaultMat.DynamicFriction = 0.5f;

            //tank size
            tankSize = new Vector3(4.1f, 3.65f, 3.4f);

            HardwareVersion ver = Core.HardwareVersion;

            //connect to the remote debugger if its there
            core.Foundation.RemoteDebugger.Connect("localhost");

            toggleMarkSleepObjecs();

            //tanks (enemies)
            this.enemyTanks = new Dictionary<long, Tank>();
            this.enemyActorsList = new Dictionary<long, Actor>();

            //collisions
            //collisionOccured = false;

            //bullets 
            bulletShotted = false;
            bulletsList = new Dictionary<int, Actor>();
            bulletsDestroyList = new Dictionary<int, double>();
            
            this.IsMouseVisible = true;

            //tank movement - rotation
            _modelRotation = Quaternion.Identity;

            base.Initialize();            
        }

        void SetActorCollisionGroup(Actor actor, short group)
        {           
            int nbShapes = actor.Shapes.Count;            

            while (Convert.ToBoolean(nbShapes--))
            {
                actor.Shapes[nbShapes].Group = group; 
            }
        }


        protected override void  LoadContent()
        {
            //foreach (var arg in parameters)
            //{
            //    Console.WriteLine(arg);
            //}
            CreateGroundPlane();
            //SelectedActor = createSphere(new Vector3(100.0f, 40.0f, 200.0f), 1.25f, 20.0f);

            box0 = createBox(new Vector3(10.0f, 0.0f, 0.0f), new Vector3(3.0f, 4.0f, 5.0f), 20.0f);
            SelectedActor = box0;

            //multiplayer
            if (multiplayer == true)
            {
                CreateObsticles();
                CreateWalls();
            }

            //single player
            if (multiplayer == false)
            {
                CreateCloth(new Vector3(0.0f, 40.0f, 0.0f));
                CreateCloth(new Vector3(50.0f, 40.0f, 50.0f));
                CreateCloth(new Vector3(-50.0f, 40.0f, -50.0f));
                // Something to drop on it
                {
                    var actorDesc = new ActorDescription()
                    {
                        GlobalPose = Matrix.CreateTranslation(0, 110, 0),
                        Shapes = { new SphereShapeDescription(2) },
                        BodyDescription = new BodyDescription(50),
                    };

                    var actor = this.Scene.CreateActor(actorDesc);
                }
                {
                    var actorDesc = new ActorDescription()
                    {
                        GlobalPose = Matrix.CreateTranslation(0, 150, 0),
                        Shapes = { new SphereShapeDescription(2) },
                        BodyDescription = new BodyDescription(50),
                    };

                    var actor = this.Scene.CreateActor(actorDesc);
                }
                {
                    var actorDesc = new ActorDescription()
                    {
                        GlobalPose = Matrix.CreateTranslation(52, 110, 53),
                        Shapes = { new SphereShapeDescription(2) },
                        BodyDescription = new BodyDescription(50),
                    };

                    var actor = this.Scene.CreateActor(actorDesc);
                }
                {
                    var actorDesc = new ActorDescription()
                    {
                        GlobalPose = Matrix.CreateTranslation(-51, 110, -51),
                        Shapes = { new SphereShapeDescription(2) },
                        BodyDescription = new BodyDescription(50),
                    };

                    var actor = this.Scene.CreateActor(actorDesc);
                }

                Tank newTank = new Tank();
                SelectedActor = createBox(new Vector3(60.0f, 0.0f, 0.0f), new Vector3(6.0f, 4.5f, 9.0f), 0.01f);
                tank = SelectedActor;
                tank.Name = "tank";
            }          

            SelectedActor.RaiseBodyFlag(BodyFlag.EnergySleepTest);

            //LoadVehicle();

            #region Load Models, Pictures & Textures

            //skybox
            effect = Content.Load<Effect>(@"Skybox\\effects");
            skyboxModel = LoadModel(@"Skybox\\skybox", out skyboxTextures);

            //plane
            planeTexture = Content.Load<Texture2D>(@"Textures\\downhback");

            //health bar
            lifeBarOrigin = Content.Load<Texture2D>(@"Pictures\\HealthBar");

            //reload bar
            reloadBarOrigin = Content.Load<Texture2D>(@"Pictures\\HealthBar");

            //font
            font = Content.Load<SpriteFont>(@"Fonts\\GameFont");

            try
            {
                _box = Content.Load<Model>(@"Models\\crate");
            }
            catch
            {
                Debug.WriteLine("Cube doesnt exists");
            }
            try
            {
                _tank = Content.Load<Model>(@"Models\\tank");
            }
            catch
            {
                Debug.WriteLine("Crate doesnt exists");
            }

            try
            {
                _capsule = Content.Load<Model>(@"Models\\capsule");
            }
            catch
            {
                Debug.WriteLine("Capsule doesnt exists");
            }
            try
            {
                _plane = Content.Load<Model>(@"Models\\plane");
            }
            catch
            {
                Debug.WriteLine("Plane doesnt exists");
            }
            try
            {
                _sphere = Content.Load<Model>(@"Models\\sphere");
            }
            catch
            {
                Debug.WriteLine("Sphere doesnt exists");
            }

            #endregion

            #region Load Sounds
            
            soundExplosion = Content.Load<SoundEffect>(@"Audio\\Explo_Hallway1");
            soundReload = Content.Load<SoundEffect>(@"Audio\\reload");
            soundFire = Content.Load<SoundEffect>(@"Audio\\tank_shot_cut");
            soundExplosionInstance = soundExplosion.CreateInstance();            
            soundReloadInstance = soundReload.CreateInstance();            
            soundFireInstance = soundFire.CreateInstance();

            //game music
            gameMusic = Content.Load<Song>(@"Audio\\Cut_and_Run");

            #endregion

            FontPos = new Vector2(GraphicsDevice.Viewport.Width / 2, GraphicsDevice.Viewport.Height / 2);
        }

        private void GenerateRandomBox()
        {
            var random = new Random();

            int w = random.Next(1, 8);
            int h = random.Next(1, 8);
            int d = random.Next(1, 8);

            int x = random.Next(-100, -80);
            int y = random.Next(-100, -80);

            var desc = new ActorDescription(new BoxShapeDescription(w, h, d))
            {
                BodyDescription = new BodyDescription(100),
                GlobalPose =
                    Matrix.CreateRotationY((float)(random.NextDouble() * 2 * Math.PI)) *
                    Matrix.CreateTranslation(x, 50, y)
            };
            var actor = this.Scene.CreateActor(desc);
        }


		private void CreateCloth(Vector3 position)
		{
			int w = 25;
			int h = 25;

			float hw = w / 2.0f;
			float hh = h / 2.0f;

            Vector3 p = position;

			// Create a Grid of Points
			int vertices, indices;

			ClothMesh clothMesh;
			{
				var grid = VertexGrid.CreateGrid(w, h);

				vertices = grid.Points.Length;
				indices = grid.Indices.Length;

				ClothMeshDescription clothMeshDesc = new ClothMeshDescription();
				clothMeshDesc.AllocateVertices<Vector3>(vertices);
				clothMeshDesc.AllocateTriangles<int>(indices / 3);

				clothMeshDesc.VertexCount = vertices;
				clothMeshDesc.TriangleCount = indices / 3;

				clothMeshDesc.VerticesStream.SetData(grid.Points);
				clothMeshDesc.TriangleStream.SetData(grid.Indices);

				//uses 32 bit integers for indices, make sure the 16 bit flag is removed.
				//32 bits are the default, so this isn't technically needed, but it's good to show in a sample
				clothMeshDesc.Flags &= ~MeshFlag.Indices16Bit;
                clothMeshDesc.Flags |= (MeshFlag)((int)clothMeshDesc.Flags | (int)ClothMeshFlag.Tearable);

				// Write the cooked data to memory
				using (var memoryStream = new MemoryStream())
				{
					Cooking.InitializeCooking();
					Cooking.CookClothMesh(clothMeshDesc, memoryStream);
					Cooking.CloseCooking();

					// Need to reset the position of the stream to the beginning
					memoryStream.Position = 0;

					clothMesh = this.Scene.Core.CreateClothMesh(memoryStream);
				}
			}

			int j = vertices * 2;
			int k = indices * 3;

			var clothDesc = new ClothDescription()
			{
				ClothMesh = clothMesh,
				GlobalPose = Matrix.CreateTranslation(-hw, 0, -hh) * Matrix.CreateTranslation(p),
                Flags = ClothFlag.Gravity | ClothFlag.Bending | ClothFlag.CollisionTwoway | ClothFlag.Visualization | ClothFlag.Hardware | ClothFlag.Tearable,
				BendingStiffness = 0.1f,
				TearFactor = 1.5f,                
			};            
			clothDesc.MeshData.AllocatePositions<Vector3>(j);
			clothDesc.MeshData.AllocateIndices<int>(k);
			clothDesc.MeshData.AllocateNormals<Vector3>(j);

			clothDesc.MeshData.MaximumVertices = j;
			clothDesc.MeshData.MaximumIndices = k;       
           
            _cloth = this.Scene.CreateCloth(clothDesc);

			// Four corner boxes to hold it in place
			var positions = new[]
			{
				new Vector3(0, 0, -hh), // Back
				new Vector3(0, 0, hh), // Front
				new Vector3(-hw, 0, 0), // Left
				new Vector3(hw, 0, 0), // Right
			};            
			var sizes = new[]
			{
				new Vector3(w, 1, 1), // Back
				new Vector3(w, 1, 1), // Front
				new Vector3(1, 1, h), // Left
				new Vector3(1, 1, h), //Right
			};

            for (int i = 0; i < 4; i++)
            {
                var actorDesc = new ActorDescription()
                {
                    GlobalPose = Matrix.CreateTranslation(positions[i] + position),
                    Shapes = { new BoxShapeDescription(sizes[i]) }
                };

                var actor = this.Scene.CreateActor(actorDesc);

                _cloth.AttachToShape(actor.Shapes.First(), (ClothAttachmentFlag.Tearable));
            }
		}

        private void CreateObsticles()
        {
            for (int i = 0; i < 15; i++)
            {
                Actor box = createBox(new Vector3(0.0f, 0.0f, -420.0f + i * 60), new Vector3(3.0f, 7.0f, 15.0f), 0);
                SetActorCollisionGroup(box, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
            }

            for (int i = 0; i < 3; i++)
            {
                Actor box1 = createBox(new Vector3(200.0f, 0.0f, -60.0f + i * 60), new Vector3(20.0f, 20.0f, 20.0f), 0.0f);
                Actor box2 = createBox(new Vector3(-200.0f, 0.0f, -60.0f + i * 60), new Vector3(20.0f, 20.0f, 20.0f), 0.0f);
                SetActorCollisionGroup(box1, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
                SetActorCollisionGroup(box2, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
            }
        }

        private void CreateWalls()
        {
            int i;
            uint numBoxes = 30;
            float halfBoxSize = 10;
            float wallLength = numBoxes * halfBoxSize * 3;
            for (i = 0; i < numBoxes; i++)
            {
                Actor box = createBox(new Vector3(((float)i / (float)(numBoxes - 1)) * wallLength - wallLength * 0.5f, 0, wallLength * 0.5f), new Vector3(halfBoxSize, halfBoxSize, halfBoxSize), 0);
                SetActorCollisionGroup(box, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
            }
            for (i = 0; i < numBoxes; i++)
            {
                Actor box = createBox(new Vector3(((float)(i) / (float)(numBoxes - 1)) * wallLength - wallLength * 0.5f, 0, -wallLength * 0.5f), new Vector3(halfBoxSize, halfBoxSize, halfBoxSize), 0);
                SetActorCollisionGroup(box, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
            }
            for (i = 0; i < numBoxes; i++)
            {
                Actor box = createBox(new Vector3(-wallLength * 0.5f, 0, ((float)(i) / (float)(numBoxes - 1)) * wallLength - wallLength * 0.5f), new Vector3(halfBoxSize, halfBoxSize, halfBoxSize), 0);
                SetActorCollisionGroup(box, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
            }
            for (i = 0; i < numBoxes; i++)
            {
                Actor box = createBox(new Vector3(wallLength * 0.5f, 0, ((float)(i) / (float)(numBoxes - 1)) * wallLength - wallLength * 0.5f), new Vector3(halfBoxSize, halfBoxSize, halfBoxSize), 0);
                SetActorCollisionGroup(box, (short)CollisionGroup.GROUP_COLLIDABLE_NON_PUSHABLE);
            }
            return;
        }


        private Model LoadModel(string assetName, out Texture2D[] textures)
        {

            Model newModel = Content.Load<Model>(assetName);
            textures = new Texture2D[newModel.Meshes.Count];
            int i = 0;
            foreach (ModelMesh mesh in newModel.Meshes)
                foreach (BasicEffect currentEffect in mesh.Effects)
                    textures[i++] = currentEffect.Texture;

            foreach (ModelMesh mesh in newModel.Meshes)
                foreach (ModelMeshPart meshPart in mesh.MeshParts)
                    meshPart.Effect = effect.Clone(this.GraphicsDevice);

            return newModel;
        }

        public void togglePause()
        {
            switch (CurrentState.Instance.GameState)
            {
                case GameState.InMenu:
                    CurrentState.Instance.GameState = GameState.Running;
                    InputManager.Instace.UpdatePreviousKeyBoardState(Keyboard.GetState());
                    Camera.CameraYaw = cameraYaw;
                    Camera.CameraPitch = cameraPitch;
                    this.IsMouseVisible = false;
                    break;
                case GameState.Running:
                    cameraYaw = Camera.CameraYaw;
                    cameraPitch = Camera.CameraPitch;
                    CurrentState.Instance.GameState = GameState.InMenu;
                    InputManager.Instace.UpdatePreviousKeyBoardState(Keyboard.GetState());
                    this.IsMouseVisible = true;
                    break;
                default:
                    break;
            }
        }

        Actor CreateGroundPlane()
        {
            // Create a plane with default descriptor
            PlaneShapeDescription planeDesc = new PlaneShapeDescription();
            ActorDescription actorDesc = new ActorDescription();
            Material defaultMaterial = Scene.Materials[0];
            planeDesc.Material = defaultMaterial;
            actorDesc.Shapes.Add(planeDesc);
            return Scene.CreateActor(actorDesc);
        }

        private Actor createSphere(Vector3 position, float radius, float density)
        {
            // Add a single-shape actor to the scene
            ActorDescription actorDesc = new ActorDescription();
            BodyDescription bodyDesc = new BodyDescription();

            // The actor has one shape, a sphere
            SphereShapeDescription sphereDesc = new SphereShapeDescription(radius);
            sphereDesc.LocalPose = Matrix.CreateTranslation(position);
            sphereDesc.SkinWidth = 0.05f;
            actorDesc.Shapes.Add(sphereDesc);

            actorDesc.BodyDescription = bodyDesc;
            actorDesc.Density = density;
            //actorDesc.GlobalPose = Matrix.CreateTranslation(position.X, position.Y + radius, position.Z);
            actorDesc.GlobalPose = Matrix.CreateTranslation(0.0f, position.Y + radius, 0.0f);

            Actor pActor = Scene.CreateActor(actorDesc);

            return pActor;
        }

        private Actor createCapsule(Vector3 position, float radius, float height, float density)
        {
            ActorDescription actorDesc = new ActorDescription();
            BodyDescription bodyDesc = new BodyDescription();

            CapsuleShapeDescription capsuleDesc = new CapsuleShapeDescription()
            {
                Radius = radius,
                Height = height,
                LocalPose = Matrix.CreateTranslation(position),
                LocalRotation = Matrix.CreateRotationZ(0),
            };

            actorDesc.Shapes.Add(capsuleDesc);
            actorDesc.BodyDescription = bodyDesc;
            actorDesc.Density = density;
            actorDesc.GlobalPose = Matrix.CreateTranslation(6.0f, capsuleDesc.Height, 0);

            if (!actorDesc.IsValid())
                throw new InvalidDataException();

            Actor pActor = Scene.CreateActor(actorDesc);

            return pActor;
        }

        private Actor createBox(Vector3 pos, Vector3 boxDim, float density)
        {
            // Add a single-shape actor to the scene
            ActorDescription actorDesc = new ActorDescription();
            BodyDescription bodyDesc = new BodyDescription();

            // The actor has one shape, a box
            BoxShapeDescription boxShapeDesc = new BoxShapeDescription();
            boxShapeDesc.Dimensions = new Vector3(boxDim.X, boxDim.Y, boxDim.Z);
            boxShapeDesc.LocalPose = Matrix.CreateTranslation(0, boxDim.Y, 0);           

            actorDesc.Shapes.Add(boxShapeDesc);
            if (density > 0)
            {
                actorDesc.BodyDescription = bodyDesc;
                actorDesc.Density = density;
            }
            actorDesc.GlobalPose = Matrix.CreateTranslation(pos);

            if (!actorDesc.IsValid())
                throw new InvalidDataException();

            Actor pActor = Scene.CreateActor(actorDesc);

            return pActor;
        }

        private Actor CreateCab(Vector3 pos)
        {
            Actor actor = createBox(pos + new Vector3(0.0f, 1.0f, 0.0f), new Vector3(1.0f, 0.5f, 0.5f), 10);
            actor.RaiseBodyFlag(BodyFlag.FrozenRotationX);
            actor.RaiseBodyFlag(BodyFlag.FrozenRotationZ);
            actor.AngularDamping = 0.5f;

            WheelShapeDescription wheelShapeDesc = new WheelShapeDescription();
            wheelShapeDesc.LocalPosition = new Vector3(0.0f, -0.4f, 0.0f);
            wheelShapeDesc.Radius = 0.5f;
            wheelShapeDesc.SkinWidth = 0.5f;

            return actor;
        }

        protected override void UnloadContent()
        {            
            for (int i = 0; i < Scene.Actors.Count(); i++)
            {                        
                Scene.Actors[i].Dispose();
            } 
        }

        public void startGame()
        {
            if (gameStarted)
                return;
            togglePause();
            gameStarted = true;
            if (multiplayer == true)
                MediaPlayer.Play(gameMusic);
        }

        protected override void Update(GameTime gameTime)
        {            
            //FPS
            elapsedTime += gameTime.ElapsedGameTime;
            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                frameRate = frameCounter;
                frameCounter = 0;
            }

            elapsedTimeFloat = gameTime.ElapsedGameTime.Milliseconds;
            elapsedTotalTime = gameTime.TotalGameTime.TotalMilliseconds;
            //game starts after 2 Update() calls
            if (gameTime.TotalGameTime.Milliseconds == 33)
            {
                startGame();
            }            
            InputManager.Instace.Update();            

            switch (CurrentState.Instance.GameState)
            {
                case GameState.Running:
                    // Update Physics
                    if (multiplayer == true)
                    {
                        UpdateClient();
                        UpdateBars();
                        UpdateSound();
                    }
                    Scene.Simulate((float)gameTime.ElapsedGameTime.TotalMilliseconds / 1000.0f);                    
                    Scene.FlushStream();
                    Scene.FetchResults(SimulationStatus.AllFinished, true);
                    Camera.Update(gameTime);
                    UpdateShooting(gameTime);                                         
                    break;
                case GameState.InMenu:
                    break;
                default:
                    break;
            }            
            base.Update(gameTime);
        }

        private void UpdateBars()
        {
            if (reloadProgress < 90 && reloadActive == true)
            {
                reloadProgress += 1;
            }            
            if (reloadProgress == 90)
            {
                reloadActive = false;
                reloadProgress = 0;
            }
            lifeBar = GenerateBar(myTank.HitPoints, 100, Color.Red);
            reloadBar = GenerateBar(reloadProgress, 94, Color.Blue);
        }


        private void UpdateSound()
        {
            if (soundFireInstance.State == SoundState.Stopped && reloadActivate == true)
            {        
                if (soundMuted == false)
                    soundReloadInstance.Play();
                reloadActive = true;
                reloadActivate = false;
            }
        }

        private void UpdateShooting(GameTime gameTime)
        {
            int bulletNumber = -1;
            long enemyID = CollisionWithEnemy(out bulletNumber);
            if (enemyID != -1 && bulletNumber != -1)
            {
                collisionCountdown = gameTime.TotalGameTime.TotalMilliseconds + 1000;
                if (!bulletsDestroyList.ContainsKey(bulletNumber))
                {                    
                    bulletsDestroyList.Add(bulletNumber, collisionCountdown);
                    enemyHit = true;
                }
            }

            //clear bullet after collision + 1000 miliseconds (collision simulation)
            foreach (KeyValuePair<int, double> bulletToDestroy in bulletsDestroyList)
            {
                if (bulletToDestroy.Value <= gameTime.TotalGameTime.TotalMilliseconds)
                {
                    if (bulletsList.ContainsKey(bulletToDestroy.Key))
                    {
                        bulletsList[bulletToDestroy.Key].Dispose();
                        bulletsList[bulletToDestroy.Key] = null;
                    }
                }
            }

            int bulletToDestroyKey = -1;
            foreach (KeyValuePair<int, Actor> bullet in bulletsList)
            {
                if (bullet.Value == null)
                {
                    bulletsDestroyList.Remove(bullet.Key);
                    bulletToDestroyKey = bullet.Key;                    
                }                   
            }

            if (bulletToDestroyKey != -1)
            {
                bulletsList.Remove(bulletToDestroyKey);
            }

            //shots are within tank -> activate collision after bullets flies out
            if ((activeTime <= gameTime.TotalGameTime.TotalMilliseconds) && bulletShotted)
            {
                if (bullet.IsDisposed != true)
                {
                    bullet.ClearActorFlag(ActorFlag.DisableCollision);
                    bulletShotted = false;                    
                }
            }
        }


        private void UpdateClient()
        {
            //send message
            NetOutgoingMessage omsg = client.CreateMessage();

            if (destroyEnemyBullet == true)
            {
                omsg.Write((int)messageType.BULLET_DESTROYED);
                destroyEnemyBullet = false;
            }
            else if (enemyHit == true)
            {
                omsg.Write((int)messageType.HIT);
                enemyHit = false;
            }
            else
            {
                omsg.Write((int)messageType.POSITION);
            }

            //my tank position
            if (tank == null)
            {
                WriteMatrix(omsg, new Matrix());
            }
            else
            {                
                WriteMatrix(omsg, tank.GlobalPose);                
            }

            //number of bullets 0 || 1
            omsg.Write(bulletsList.Count);
         
            if (bulletsList.Count == 0)
            {
                WriteMatrix(omsg, new Matrix());                
            }
            else
            {
                //WriteMatrix(omsg, bulletsList[0].GlobalPose);
                foreach (KeyValuePair<int, Actor> bullet in bulletsList)
                {
                    WriteMatrix(omsg, bullet.Value.GlobalPose);
                }
            }            
            client.SendMessage(omsg, NetDeliveryMethod.ReliableOrdered);

        ////////////////////////////////////////////////////////////////////////////
            //read message
            NetIncomingMessage imsg;
            while ((imsg = client.ReadMessage()) != null)
            {
                switch (imsg.MessageType)
                {
                    //discovery response from server
                    case NetIncomingMessageType.DiscoveryResponse:
                        string serverhello = imsg.ReadString();
                        client.Connect(imsg.SenderEndpoint);
                        playerID = client.UniqueIdentifier;

                        //Tank newTank = new Tank();
                        if (serverhello.Equals("firstPlayer"))
                        {
                            tank = createBox(new Vector3(60.0f, 0.0f, 0.0f), new Vector3(6.0f, 4.5f, 9.0f), 0.01f);
                        }
                        else if (serverhello.Equals("secondPlayer"))
                            tank = createBox(new Vector3(-60.0f, 0.0f, 0.0f), new Vector3(6.0f, 4.5f, 9.0f), 0.01f);

                        tank.Name = "tank";
                        SelectedActor = tank;
                        break;

                    //getting data
                    case NetIncomingMessageType.Data:  
                        //clear list containing tank & bullet pose                   
                        positionsMatrixes.Clear();

                        long who = imsg.ReadInt64();

                        //message identification
                        Int32 messageID = (int)imsg.ReadSingle();

                        if (((messageType)messageID) == messageType.BULLET_DESTROYED)
                        {
                            if (enemyBulletActor != null)
                            {
                                enemyBulletActor.Dispose();
                                enemyBulletActor = null;
                            }
                        }
                        else if (((messageType)messageID) == messageType.HIT && playerID != who)
                        {
                            myTank.Hit(25); 
                        }
                       
                        //tank pose - my + enemyTank 
                        ReadMatrix(imsg, ref matrixPose);
                        positionsMatrixes.Insert(0, matrixPose);

                        //bullets number
                        int bulletsCountDebug = (int)imsg.ReadSingle();

                        //bullet pose
                        ReadMatrix(imsg, ref matrixBulletPose);                                                
                        positionsMatrixes.Insert(1, matrixBulletPose);

                        positions2[who] = positionsMatrixes;

                        if (playerID != who)
                        {
                            if (enemyActor == null)
                            {
                                Tank newEnemyTank = new Tank();
                                if (tank.GlobalPosition.X > 50.0f && tank.GlobalPosition.X < 70.0f)
                                    enemyActor = createBox(new Vector3(-20.0f, 0.0f, 0.0f), new Vector3(6.0f, 4.5f, 9.0f), 0.01f);
                                else
                                    enemyActor = createBox(new Vector3(20.0f, 0.0f, 0.0f), new Vector3(6.0f, 4.5f, 9.0f), 0.01f);
                                    
                                enemyActor.Name = "enemy";
                                this.enemyActorsList.Add(who, enemyActor);
                                this.enemyTanks.Add(who, newEnemyTank);
                            }
                            this.enemyTanks[who].Pose = positions2[who][0];
                            this.enemyActorsList[who].GlobalPose = this.enemyTanks[who].Pose;

                            //bullets                            
                            if (bulletsCountDebug == 1)
                            {
                                if (enemyBulletActor == null)
                                {
                                    enemyBulletActor = createSphere(new Vector3(enemyActorsList[who].GlobalPosition.X, enemyActorsList[who].GlobalPosition.Y + 3.5f, enemyActorsList[who].GlobalPosition.Z), 0.8f, 20.0f);                                    
                                }                                
                            }     
                            else if (enemyBulletActor != null)
                            {
                                enemyBulletActor.Dispose();
                                enemyBulletActor = null;
                            }
                            if (enemyBulletActor != null)
                            {
                                enemyBulletActor.GlobalPose = positions2[who][1];                             
                            }
                        }
                        break;
                }
            }
        }


        private long CollisionWithEnemy(out int bulletNumber)
        {
            if (bulletsList.Count == 0)
            {
                bulletNumber = -1;
                return -1;
            }
            foreach (KeyValuePair<int, Actor> bullet in this.bulletsList)
            {
                foreach (SphereShape sphereShape in bullet.Value.Shapes)
                {
                    foreach (ModelMesh sphereMesh in _sphere.Meshes)
                    {
                        foreach (KeyValuePair<long, Actor> tank in this.enemyActorsList)
                        {
                            foreach (BoxShape boxShape in tank.Value.Shapes)
                            {
                                foreach (ModelMesh tankMesh in _tank.Meshes)
                                {
                                    BoundingSphere tankBoundingSphere = tankMesh.BoundingSphere;
                                    tankBoundingSphere.Center = boxShape.GlobalPose.Translation;
                                    tankBoundingSphere.Radius = 6.0f;

                                    BoundingSphere sphereBoundingSphere = sphereMesh.BoundingSphere;
                                    sphereBoundingSphere.Center = sphereShape.GlobalPose.Translation;
                                    sphereBoundingSphere.Radius = 5.5f;

                                    if (sphereBoundingSphere.Intersects(tankBoundingSphere))
                                    {
                                        if (soundMuted == false)
                                            soundExplosionInstance.Play();
                                        bulletNumber = bullet.Key;                                        
                                        return tank.Key;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            bulletNumber = -1;
            return -1;
        }            

        public void toggleMarkSleepObjecs()
        {
            markSpleepObjects = !markSpleepObjects;
        }

        private Vector3 ApplyForceToActor(Actor actor, Vector3 forceDir, float gForceStrenght)
        {
            Vector3 forceVec = gForceStrenght * forceDir * elapsedTimeFloat * 0.001f;
            //limitation of tank speed
            if (actor.Name != null && (actor.Name.Equals("tank") || actor.Name.Equals("enemy")))
            {
                if (actor.LinearMomentum.X > 1000.0f)
                    actor.LinearMomentum = new Vector3(1000.0f, actor.LinearMomentum.Y, actor.LinearMomentum.Z);
                if (actor.LinearMomentum.Y > 1000.0f)
                    actor.LinearMomentum = new Vector3(actor.LinearMomentum.X, 1000.0f, actor.LinearMomentum.Z);
                if (actor.LinearMomentum.Z > 1000.0f)
                    actor.LinearMomentum = new Vector3(actor.LinearMomentum.X, actor.LinearMomentum.Y, 1000.0f);

                if (actor.LinearMomentum.X < -1000.0f)
                    actor.LinearMomentum = new Vector3(-1000.0f, actor.LinearMomentum.Y, actor.LinearMomentum.Z);
                if (actor.LinearMomentum.Y < -1000.0f)
                    actor.LinearMomentum = new Vector3(actor.LinearMomentum.Y, -1000.0f, actor.LinearMomentum.Z);
                if (actor.LinearMomentum.Z < -1000.0f)
                    actor.LinearMomentum = new Vector3(actor.LinearMomentum.X, actor.LinearMomentum.Y, -1000.0f);
            }
            actor.AddForce(forceVec, ForceMode.Force);
            return forceVec;
        }

        private void selectNextActor()
        {
            Actor actor;
            for (int i = 0; i < Scene.Actors.Count; i++)
            {
                actor = Scene.Actors[i];
                if (actor.Equals(gSelectedActor))
                {
                    uint j = 1;
                    gSelectedActor = Scene.Actors[(int)((i + j) % Scene.Actors.Count)];
                    while (!IsSelectable(gSelectedActor))
                    {
                        j++;
                        gSelectedActor = Scene.Actors[(int)((i + j) % Scene.Actors.Count)];
                    }
                    break;
                }

            }
        }

        private bool IsSelectable(Actor actor)
        {
            foreach (Shape shape in gSelectedActor.Shapes)
            {
                if (shape.GetFlag(ShapeFlag.TriggerEnable))
                    return false;
            }
            if (multiplayer == true)
            {
                if (!actor.IsDynamic)
                    return false;
                if (actor.Equals(groundPlane))
                    return false;
                if (actor.Name == null)
                    return false;
                if (!actor.Name.Equals("tank"))
                    return false;
            }
            else
            {
                if (!actor.IsDynamic)
                    return false;
            }
            return true;
        }

        public Texture2D GenerateBar(int Current, int Max, Color fillColor)
        {
                #region Vars
                int BarWidth = 50;
                int BarHeight = 5;
                int BoarderOffSet = 2;
                Double PercentToDraw = (Double)Current / Max;
                Double EachPercentWidth = (Double)BarWidth / Max;
                #endregion


                #region rectangle setup
                //Boarder Rectangle
                Rectangle RecBoarder = new Rectangle(0, 0, BarWidth + (BoarderOffSet * 2), BarHeight + (BoarderOffSet * 2));

                //Bar Background Rectangle
                Rectangle RecBackGround = new Rectangle(BoarderOffSet, BoarderOffSet, BarWidth, BarHeight);

                //Fill Rectangle
                Rectangle RecFill = new Rectangle(BoarderOffSet, BoarderOffSet, (int)((PercentToDraw * 100) * EachPercentWidth), BarHeight);
                #endregion


                #region draw the bar
                GraphicsDevice device = this.GraphicsDevice;

                RenderTarget2D renderTarget = new RenderTarget2D(device, RecBoarder.Width,
                RecBoarder.Height, 1, device.DisplayMode.Format, device.PresentationParameters.MultiSampleType, device.PresentationParameters.MultiSampleQuality, RenderTargetUsage.DiscardContents);

                device.SetRenderTarget(0, renderTarget);
                spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.None);

                spriteBatch.Draw(lifeBarOrigin, RecBoarder, Color.Black);
                spriteBatch.Draw(lifeBarOrigin, RecBackGround, Color.Gray);
                spriteBatch.Draw(lifeBarOrigin, RecFill, fillColor);
                spriteBatch.End();
                device.SetRenderTarget(0, null);
                #endregion
                return renderTarget.GetTexture();
        }


        protected override void Draw(GameTime gameTime)
        {
            //FPS
            frameCounter++;

            this.GraphicsDevice.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, Color.Black, 1.0f, 0);

            DrawSkybox();

            GraphicsDevice.VertexDeclaration = new VertexDeclaration(GraphicsDevice, VertexPositionColor.VertexElements);

            VisualizationEffect.World = Matrix.Identity;
            VisualizationEffect.View = Camera.View;
            VisualizationEffect.Projection = Camera.Projection;

            DebugRenderable data = Scene.GetDebugRenderable();

            if (bIsDebug)
            {
                VisualizationEffect.Begin();

                foreach (EffectPass pass in VisualizationEffect.CurrentTechnique.Passes)
                {
                    pass.Begin();

                    if (data.PointCount > 0)
                    {
                        DebugPoint[] points = data.GetDebugPoints();
                        this.GraphicsDevice.DrawUserPrimitives<DebugPoint>(PrimitiveType.PointList, points, 0, points.Length);                                            
                    }

                    if (data.LineCount > 0)
                    {
                        DebugLine[] lines = data.GetDebugLines();

                        VertexPositionColor[] vertices = new VertexPositionColor[data.LineCount * 2];
                        for (int x = 0; x < data.LineCount; x++)
                        {
                            DebugLine line = lines[x];

                            vertices[x * 2 + 0] = new VertexPositionColor(line.Point0, Int32ToColor(line.Color));
                            vertices[x * 2 + 1] = new VertexPositionColor(line.Point1, Int32ToColor(line.Color));
                        }
                        if (debug == true)
                            this.GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, vertices, 0, lines.Length);
                    }

                    if (data.TriangleCount > 0)
                    {
                        DebugTriangle[] triangles = data.GetDebugTriangles();

                        VertexPositionColor[] vertices = new VertexPositionColor[data.TriangleCount * 3];
                        for (int x = 0; x < data.TriangleCount; x++)
                        {
                            DebugTriangle triangle = triangles[x];

                            vertices[x * 3 + 0] = new VertexPositionColor(triangle.Point0, Int32ToColor(triangle.Color));
                            vertices[x * 3 + 1] = new VertexPositionColor(triangle.Point1, Int32ToColor(triangle.Color));
                            vertices[x * 3 + 2] = new VertexPositionColor(triangle.Point2, Int32ToColor(triangle.Color));
                        }
                        this.GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.TriangleList, vertices, 0, triangles.Length);
                    }
                    pass.End();
                }
                VisualizationEffect.End();
            }
            renderActors();
            gForceVec = new Vector3(0, 10.0f, 0);

            Vector2 posLeft = new Vector2(0, 0);
            Vector2 posRight = new Vector2(deviceManager.GraphicsDevice.Viewport.Width, 0);
            float posMiddleY = deviceManager.GraphicsDevice.Viewport.Height / 2;
            float posMiddleX = deviceManager.GraphicsDevice.Viewport.Width / 2;

            if (CurrentState.Instance.GameState == GameState.InMenu)
            {
                spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
                string output = "Paused - press 'P' to resume";

                // Find the center of the string
                Vector2 FontOrigin = font.MeasureString(output) / 2;
                // Draw the string
                spriteBatch.DrawString(font, output, FontPos, Color.Red,
                    0, FontOrigin, 1.0f, SpriteEffects.None, 1.0f);
                spriteBatch.End();
            }
            else
            {
                spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);

                Vector2 FontPosition = new Vector2(100.0f, 0.0f);

                //network game - 2 players connected
                if (positions2.Count == 2)
                {
                    if (myTank.HitPoints == 0)
                    {
                        string output3 = "GAME OVER";
                        Vector2 FontOrigin3 = font.MeasureString(output3) / 2;
                        spriteBatch.DrawString(font, output3, new Vector2(posMiddleX, posMiddleY - 20), Color.Red,
                                               0, FontOrigin3, 1.5f, SpriteEffects.None, 1.0f);
                    }
                }
                else
                {
                    string fps = string.Format("FPS: {0} {1} {2} {3} {4} {5} {6} {7} {8}", frameRate, "\nNumber of actors:",
                                                Scene.Actors.Count(), "\nNumber of cloth:", Scene.Cloths.Count(),
                                                "\nvsync:", deviceManager.SynchronizeWithVerticalRetrace,
                                                "\nNumber of substeps:", Scene.TimingInformation.NumberOfSubSteps);
                    Vector2 FontOrigin1 = font.MeasureString(fps) / 2;
                    spriteBatch.DrawString(font, fps, new Vector2(140, 100), Color.Red,
                                           0, FontOrigin1, 1.0f, SpriteEffects.None, 1.0f);
                }

                if (multiplayer == true)
                {
                    //health bar
                    spriteBatch.Draw(lifeBar, new Rectangle(GraphicsDevice.Viewport.X + 150, 10, 500, 20), Color.Red);
                    //reload bar
                    if (reloadActive == true)
                        spriteBatch.Draw(reloadBar, new Rectangle(GraphicsDevice.Viewport.X + 200, 155, 400, 10), Color.Blue);
                }
                spriteBatch.End();
            }
            base.Draw(gameTime);
        }


        private void DrawSkybox()
        {
            GraphicsDevice.SamplerStates[0].AddressU = TextureAddressMode.Clamp;
            GraphicsDevice.SamplerStates[0].AddressV = TextureAddressMode.Clamp;

            this.GraphicsDevice.RenderState.DepthBufferEnable = false;

            Matrix[] skyboxTransforms = new Matrix[skyboxModel.Bones.Count];
            skyboxModel.CopyAbsoluteBoneTransformsTo(skyboxTransforms);
            int i = 0;
            foreach (ModelMesh mesh in skyboxModel.Meshes)
            {
                foreach (Effect currentEffect in mesh.Effects)
                {
                    Matrix worldMatrix = skyboxTransforms[mesh.ParentBone.Index] * Matrix.CreateScale(1000.0f);                  
                    currentEffect.CurrentTechnique = currentEffect.Techniques["Textured"];
                    currentEffect.Parameters["xWorld"].SetValue(worldMatrix);
                    currentEffect.Parameters["xView"].SetValue(Camera.View);
                    currentEffect.Parameters["xProjection"].SetValue(Camera.Projection);
                    currentEffect.Parameters["xTexture"].SetValue(skyboxTextures[i++]);
                }
                mesh.Draw();
            }
            this.GraphicsDevice.RenderState.DepthBufferEnable = true;
        }


        private void renderActors()
        {
            foreach (Actor actor in Scene.Actors)
            {
                drawActor(actor);
            }
        }

        private void drawActor(Actor actor)
        {
            Vector3 actualColor = defaultColor;
            if (actor.Equals(gSelectedActor))
                actualColor = selectedColor;

            if (isFilteringOn)
            {
                uint userData = (uint)actor.UserData;

                if (userData != 0)
                {
                    float size = userData & 0xffff;

                    uint random = userData >> 16;

                    if ((random & 1) != 0)
                        actualColor = new Vector3(1, 1, 1);
                    else
                        actualColor = new Vector3(1, 0, 0);
                }
            }

            int i = 0;

            foreach (Shape shape in actor.Shapes)
            {
                switch (shape.Type)
                {
                    case ShapeType.Box:                        
                        if (actor.Name != null && (actor.Name.Equals("tank") || actor.Name.Equals("enemy")))
                        {
                            if (actor.IsSleeping && markSpleepObjects)
                                drawTank((BoxShape)shape, tankColor);
                            else
                                drawTank((BoxShape)shape, tankColor);
                        }
                        else
                        {
                            if (actor.IsSleeping && markSpleepObjects)
                                drawBox1((BoxShape)shape, new Vector3(0, 0, 1));
                            else
                                drawBox1((BoxShape)shape, actualColor);
                        }
                        break;
                    case ShapeType.Capsule:
                        if (actor.IsSleeping && markSpleepObjects)
                            drawCapsule((CapsuleShape)shape, new Vector3(0, 0, 1));
                        else
                            drawCapsule((CapsuleShape)shape, actualColor);
                        break;
                    case ShapeType.Plane:
                        drawPlane((PlaneShape)shape);
                        break;
                    case ShapeType.Sphere:
                        if (actor.IsSleeping && markSpleepObjects)
                            drawSphere((SphereShape)shape, new Vector3(0, 0, 1));
                        else
                            drawSphere((SphereShape)shape, actualColor);
                        break;
                }
                i++;
            }
        }

        private void drawSphere(SphereShape sphereShape, Vector3 color)
        {
            // Draw the Sphere
            {
                Matrix[] transforms = new Matrix[_sphere.Bones.Count];
                _sphere.CopyAbsoluteBoneTransformsTo(transforms);                

                foreach (ModelMesh mesh in _sphere.Meshes)
                {
                    foreach (BasicEffect effect in mesh.Effects)
                    {
                        effect.DiffuseColor = color;
                        effect.EnableDefaultLighting();
                        effect.World = Matrix.CreateScale(sphereShape.Radius, sphereShape.Radius, sphereShape.Radius) * transforms[mesh.ParentBone.Index] * sphereShape.GlobalPose;
                        effect.View = Camera.View;
                        effect.Projection = Camera.Projection;
                    }
                    mesh.Draw();
                }
            }
        }

        private void drawBox1(BoxShape shape, Vector3 color)
        {
            // Draw the Box
            Matrix[] transforms = new Matrix[_box.Bones.Count];
            _box.CopyAbsoluteBoneTransformsTo(transforms);

            foreach (ModelMesh mesh in _box.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.DiffuseColor = color;
                    effect.EnableDefaultLighting();
                    effect.World = Matrix.CreateScale(new Vector3(0.01f * shape.Size.X, 0.01f * shape.Size.Y, 0.01f * shape.Size.Z)) * shape.GlobalPose;                  
                    effect.View = Camera.View;
                    effect.Projection = Camera.Projection;
                }
                mesh.Draw();
            }
        }

        private void drawTank(BoxShape shape, Vector3 color)
        {
            // Draw the Box
            Matrix[] transforms = new Matrix[_tank.Bones.Count];
            _tank.CopyAbsoluteBoneTransformsTo(transforms);

            foreach (ModelMesh mesh in _tank.Meshes)
            {
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.DiffuseColor = color;
                    effect.EnableDefaultLighting();
                    Matrix translationMatrix = Matrix.CreateTranslation(new Vector3(0.0f, 0.0f, -11.0f));
                    Matrix rotationMatrix = Matrix.CreateRotationZ(MathHelper.ToRadians(90));
                    effect.World = Matrix.CreateScale(tankSize.X * shape.Dimensions.X, tankSize.Y * shape.Dimensions.Y, tankSize.Z * shape.Dimensions.Z) * rotationMatrix * translationMatrix * transforms[mesh.ParentBone.Index] * shape.GlobalPose;                                        
                    effect.View = Camera.View;
                    effect.Projection = Camera.Projection;                    
                }
                mesh.Draw();
            }
        }

        private void drawCapsule(CapsuleShape shape, Vector3 color)
        {
            {
                Matrix[] transforms = new Matrix[_capsule.Bones.Count];
                _capsule.CopyAbsoluteBoneTransformsTo(transforms);

                foreach (ModelMesh mesh in _capsule.Meshes)
                {
                    foreach (BasicEffect effect in mesh.Effects)
                    {
                        effect.DiffuseColor = color;
                        effect.EnableDefaultLighting();

                        effect.World = Matrix.CreateScale(shape.Radius, shape.Height, shape.Radius) * transforms[mesh.ParentBone.Index] * shape.GlobalPose;
                        effect.View = Camera.View;
                        effect.Projection = Camera.Projection;
                    }
                    mesh.Draw();
                }
            }
        }

        private void drawPlane(PlaneShape shape)
        {
            {
                Matrix[] transforms = new Matrix[_plane.Bones.Count];
                _plane.CopyAbsoluteBoneTransformsTo(transforms);

                foreach (ModelMesh mesh in _plane.Meshes)
                {
                    foreach (BasicEffect effect in mesh.Effects)
                    {
                        effect.EnableDefaultLighting();
                        effect.AmbientLightColor = new Vector3(1.0f, 10.0f, 5.0f);
                        effect.World = Matrix.CreateScale(10240.0f, 1.0f, 10240.0f) * transforms[mesh.ParentBone.Index] * shape.GlobalPose;
                        effect.View = Camera.View;
                        effect.Projection = Camera.Projection;
                        effect.TextureEnabled = true;
                        effect.Texture = planeTexture;                       
                    }
                    mesh.Draw();
                }
            }

        }

        private Color Int32ToColor(int color)
        {
            byte a = (byte)((color & 0xFF000000) >> 32);
            byte r = (byte)((color & 0x00FF0000) >> 16);
            byte g = (byte)((color & 0x0000FF00) >> 8);
            byte b = (byte)((color & 0x000000FF) >> 0);

            return new Color(r, g, b, a);
        }

        private Vector3 RGBToColor(Vector3 color)
        {
            float coeficient = 1.0f / 255.0f;
            return new Vector3(coeficient * color.X, coeficient * color.Y, coeficient * color.Z);
        }

        protected override void OnExiting(object sender, EventArgs args)
        {
            client.Shutdown("bye");
            base.OnExiting(sender, args);
        }

        /// <summary>
        /// Writes a Vector3
        /// </summary>
        public static void WriteVector3(NetOutgoingMessage omsg, Vector3 vector)
        {
            omsg.Write(vector.X);
            omsg.Write(vector.Y);
            omsg.Write(vector.Z);
        }

        /// <summary>
        /// Reads a Vector3
        /// </summary>
        public static Vector3 ReadVector3(NetIncomingMessage imsg)
        {
            Vector3 retval;
            retval.X = imsg.ReadSingle();
            retval.Y = imsg.ReadSingle();
            retval.Z = imsg.ReadSingle();
            return retval;
        }

        /// <summary>
        /// Writes an orthonormal matrix (rotation, translation but no scaling or projection)
        /// </summary>
        public static void WriteMatrix(NetOutgoingMessage omsg, Matrix matrix)
        {
            omsg.Write(matrix.M11);
            omsg.Write(matrix.M12);
            omsg.Write(matrix.M13);
            omsg.Write(matrix.M14);
            omsg.Write(matrix.M21);
            omsg.Write(matrix.M22);
            omsg.Write(matrix.M23);
            omsg.Write(matrix.M24);
            omsg.Write(matrix.M31);
            omsg.Write(matrix.M32);
            omsg.Write(matrix.M33);
            omsg.Write(matrix.M34);
            omsg.Write(matrix.M41);
            omsg.Write(matrix.M42);
            omsg.Write(matrix.M43);
            omsg.Write(matrix.M44);
        }

        /// <summary>
        /// Reads a matrix written using WriteMatrix()
        /// </summary>
        public static void ReadMatrix(NetIncomingMessage imsg, ref Matrix destination)
        {
            destination.M11 = imsg.ReadSingle();
            destination.M12 = imsg.ReadSingle();
            destination.M13 = imsg.ReadSingle();
            destination.M14 = imsg.ReadSingle();
            destination.M21 = imsg.ReadSingle();
            destination.M22 = imsg.ReadSingle();
            destination.M23 = imsg.ReadSingle();
            destination.M24 = imsg.ReadSingle();
            destination.M31 = imsg.ReadSingle();
            destination.M32 = imsg.ReadSingle();
            destination.M33 = imsg.ReadSingle();
            destination.M34 = imsg.ReadSingle();
            destination.M41 = imsg.ReadSingle();
            destination.M42 = imsg.ReadSingle();
            destination.M43 = imsg.ReadSingle();
            destination.M44 = imsg.ReadSingle();         
        }


        #region Public Properties

        public SpriteBatch SpriteBatch
        {
            get { return spriteBatch; }
        }

        public Actor SelectedActor
        {
            get { return gSelectedActor; }
            set { gSelectedActor = value; }
        }

        public static Vector3 TankPosition
        {
            get { return tank.GlobalPosition; }
        }

        public static Quaternion TankOrientationQuat
        {
            get { return tank.GlobalOrientationQuat; }
        }

        public static string[] Parameters
        {
            set { parameters = value; }
        }

        #endregion

        #region KeyboardListener Members

        public void  KeyPressed(Keys key)
        {
            if (forceKeysDisabled)
                return;
            switch (key)
            {
                case Keys.P:
                    togglePause();
                    break;
                //allows the game to exit
                case Keys.Escape:
                    this.Exit();
                    break;
                //disable/enable sounds
                case Keys.Y:
                    if (soundMuted == false)
                    {
                        MediaPlayer.Pause();
                        soundMuted = true;
                    }
                    else
                    {
                        MediaPlayer.Resume();                        
                        soundMuted = false;
                    }
                    break;
                default:
                    break;
            }
        }


        public void KeyPressedRunning(Keys key)
        {    
            if (forceKeysDisabled)
               return;            
            
            switch(key)
            {
                case Keys.R:
                    selectNextActor();
                    break;
                case Keys.T:
                    tank.GlobalOrientationQuat = Quaternion.Identity;
                    tank.GlobalPosition = new Vector3(tank.GlobalPosition.X, 0.0f, tank.GlobalPosition.Z);
                    tank.Sleep();
                    break;
                case Keys.OemTilde:
                    if (deviceManager.SynchronizeWithVerticalRetrace == false)
                        deviceManager.SynchronizeWithVerticalRetrace = true;
                    else
                        deviceManager.SynchronizeWithVerticalRetrace = false;
                    if (this.IsFixedTimeStep == false)
                        this.IsFixedTimeStep = true;
                    else
                        this.IsFixedTimeStep = false;

                    deviceManager.ApplyChanges();
                    break;
                case Keys.Space:
                    if (reloadActive == false)
                    {
                        if (soundMuted == false)
                        {
                            soundFireInstance.Volume = 0.5f;
                            soundFireInstance.Play();
                        }
                        bullet = createSphere(new Vector3(tank.GlobalPosition.X, tank.GlobalPosition.Y + 3.5f, tank.GlobalPosition.Z), 0.8f, 20.0f);    
                        bullet.RaiseActorFlag(ActorFlag.DisableCollision);
                        bullet.RaiseBodyFlag(BodyFlag.DisableGravity);
                        bullet.Name = "bullet";
      
                        gForceVec = ApplyForceToActor(bullet, newForward - newBackward, 25000000.0f);
                        
                        //only 1 bullet in multiplayer
                        if (multiplayer == true) 
                            bulletsKeys = i % 1; 
                        else //simulation mode 10 bullets available
                            bulletsKeys = i % 10; 

                        if (bulletsList.ContainsKey(bulletsKeys))
                        {
                            bulletsList[bulletsKeys].Dispose();
                            destroyEnemyBullet = true;
                        }
                        bulletsList[bulletsKeys] = bullet;
                        activeTime = elapsedTotalTime + 49;
                        bulletShotted = true;
                        reloadActivate = true;              
                        i++;
                    }                        
                    break;  
                default:
                    break;            
            }
        }

        public void KeyHoldedRunning(Keys key)
        {
            if (forceKeysDisabled)
                return;
            
            switch(key)
            {
                //selected object movement
                case Keys.J:
                    gForceVec = ApplyForceToActor(gSelectedActor, new Vector3(3, 0, 0), gForceStrenght);
                    break;
                case Keys.K:
                    gForceVec = ApplyForceToActor(gSelectedActor, new Vector3(0, 0, -3), gForceStrenght);
                    break;
                case Keys.I:
                    gForceVec = ApplyForceToActor(gSelectedActor, new Vector3(0, 0, 3), gForceStrenght);
                    break;
                case Keys.L:
                    gForceVec = ApplyForceToActor(gSelectedActor, new Vector3(-3, 0, 0), gForceStrenght);
                    break;
                case Keys.U:
                    gForceVec = ApplyForceToActor(gSelectedActor, new Vector3(0, 3, 0), gForceStrenght);
                    break;
                case Keys.M:
                    gForceVec = ApplyForceToActor(gSelectedActor, new Vector3(0, -3, 0), gForceStrenght);
                    break;                
                    
                //tank movement
                case Keys.Up:                    
                    gForceVec = ApplyForceToActor(gSelectedActor, newForward * 3, gForceStrenght);
                    break;
                case Keys.Down:
                    gForceVec = ApplyForceToActor(gSelectedActor, newBackward * 3, gForceStrenght);
                    break;
                case Keys.Left:
                    rot = Quaternion.CreateFromAxisAngle(Vector3.Up, 0.025f);
                    gSelectedActor.GlobalOrientationQuat *= rot;
                    break;
                case Keys.Right:
                    rot = Quaternion.CreateFromAxisAngle(Vector3.Up, -0.025f);
                    gSelectedActor.GlobalOrientationQuat *= rot;
                    break;
                case Keys.N:
                    GenerateRandomBox();
                    break;
                default:
                    break;
            }
            newForward = Vector3.Transform(Vector3.Forward, tank.GlobalOrientationQuat);
            newBackward = Vector3.Transform(Vector3.Backward, tank.GlobalOrientationQuat);            
        }

        public void KeyPressedInMenu(Keys key)
        {
        }

        #endregion

        #region MouseListener Members

        public void MouseMoved(MouseState mouseState)
        {
        }
 

        public void MouseLeftButtonPressed()
        {
            GenerateRandomBox();                     
        }

        public void MouseRightButtonPressed()
        {
            if (Scene.Actors[Scene.Actors.Count - 1].Name == "tank" || 
                Scene.Actors[Scene.Actors.Count - 1].Name == "bullet") 
                return;
            Scene.Actors[Scene.Actors.Count - 1].Dispose();
        }

        #endregion                 

    }
}

